from __future__ import annotations

from typing import Any, Callable

from . import params


def Depends(dependency: Callable[..., Any], *, use_cache: bool = True) -> Any:  # noqa: N802
    return params.Depends(dependency=dependency, use_cache=use_cache)



from __future__ import annotations

import asyncio
from abc import ABC, abstractmethod
from dataclasses import dataclass
from functools import wraps
from logging import Logger, getLogger
from random import uniform
from typing import Callable, Generator, Optional, Type


class RetryRule(ABC):
    def __init__(
        self,
        exception: Type[BaseException],
        retries: Optional[int] = None,
        filter: Optional[Callable[[BaseException], bool]] = None,
    ):
        self.exception = exception
        self.retries = retries
        self.filter = filter

    def __call__(self, logger: Optional[Logger] = None):
        logger = logger or getLogger(__name__)

        def logger_wrap(func):
            retry_count = self.retries
            wait_time = self.generate_wait_time()

            @wraps(func)
            async def inner(*args, **kwargs):
                nonlocal retry_count
                try:
                    return await func(*args, **kwargs)
                except self.exception as e:
                    # filter条件を満たさないエラーはraise
                    if self.filter is not None and self.filter(e) is False:
                        raise
                    # カウントが0以下ならエラーをraise
                    if retry_count is not None:
                        if retry_count <= 0:
                            raise
                        else:
                            retry_count -= 1
                    sleep = next(wait_time)
                    logger.warning(f"[Retry] {e.__class__.__name__}を検知: {sleep}秒後に再実行します (残り{retry_count}回)")
                    await asyncio.sleep(sleep)
                    return await inner(*args, **kwargs)
                except BaseException:
                    raise

            return inner

        return logger_wrap

    @abstractmethod
    def generate_wait_time(self) -> Generator[float, None, None]:
        """次の待機時間を生成するGenerator"""
        raise NotImplementedError


class RetryFixed(RetryRule):
    def __init__(
        self,
        exception: Type[BaseException],
        *,
        wait_time: float,
        retries: Optional[int] = None,
        filter: Optional[Callable[[BaseException], bool]] = None,
    ):
        self.wait_time = wait_time
        super().__init__(exception=exception, retries=retries, filter=filter)

    def generate_wait_time(self) -> Generator[float, None, None]:
        while True:
            yield self.wait_time


class RetryFibonacci(RetryRule):
    def generate_wait_time(self) -> Generator[float, None, None]:
        x, y = 0.0, 1.0
        while True:
            yield y
            x, y = y, x + y


class RetryRange(RetryRule):
    def __init__(
        self,
        exception: Type[BaseException],
        *,
        min: float,
        max: float,
        retries: Optional[int] = None,
        filter: Optional[Callable[[BaseException], bool]] = None,
    ):
        self.min = min
        self.max = max
        super().__init__(exception=exception, retries=retries, filter=filter)

    def generate_wait_time(self) -> Generator[float, None, None]:
        while True:
            yield uniform(max(0, self.min), max(0, self.max))


class RetryExponentialAndJitter(RetryRule):
    def __init__(
        self,
        exception: Type[BaseException],
        *,
        base_wait_time: float = 1.0,
        cap: float = 60 * 60,
        retries: Optional[int] = None,
        filter: Optional[Callable[[BaseException], bool]] = None,
    ):
        self.base_wait_time = base_wait_time
        self.cap = cap
        super().__init__(exception=exception, retries=retries, filter=filter)

    def generate_wait_time(self) -> Generator[float, None, None]:
        retry_count = 0
        yield self.base_wait_time
        while True:
            retry_count += 1
            yield uniform(0, min(self.cap, self.base_wait_time * (2**retry_count)))


@dataclass
class Retry:
    rules: list[RetryRule]
    logger: Logger = getLogger(__name__)

    def __call__(self, func: Callable):
        @wraps(func)
        async def generate_composite(*args, **kwargs):
            return await composite(func, [rule(self.logger) for rule in self.rules])(*args, **kwargs)

        return generate_composite


def composite(base, decorators: list[Callable]):
    if len(decorators) == 0:
        return base

    func = decorators[0]

    @func
    async def wrap(*args, **kwargs):
        return await base(*args, **kwargs)

    return composite(wrap, decorators[1:])



__version__ = "0.2.0"

import logging

logging.getLogger(__name__).addHandler(logging.NullHandler())
logging.getLogger(__name__).propagate = False

from .core.depends_contoroller import get_depends  # noqa
from .core.machina import Event, Machina  # noqa
from .core.params_function import Depends  # noqa



from __future__ import annotations

import logging

try:
    from typing import Literal  # type: ignore
except ImportError:
    from typing_extensions import Literal


def set_verbose(
    logger: logging.Logger | None = None,
    verbose: Literal["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"] | None = None,
):
    """exmachinaのログを表示する

    Args:
        verbose (Literal[): 10, 20, 30, 40, 50に相当
    """
    try:
        from rich.logging import RichHandler
    except ModuleNotFoundError:
        raise ImportError("pip install exmachina[rich]")

    if verbose is not None:
        handler = RichHandler(rich_tracebacks=True, enable_link_path=False, level=verbose)
        if logger is None:
            logger = logging.getLogger("exmachina")
        logger.setLevel(verbose)
        if "RichHandler" not in [h.__class__.__name__ for h in logger.handlers]:
            logger.addHandler(handler)



from __future__ import annotations

import inspect
from typing import Any, AsyncGenerator, Callable, Generator

from . import exception as E
from .params import Depends


class DependsContoroller:
    generators: list[AsyncGenerator | Generator] = []
    caches: dict[Callable, Any] = {}

    @classmethod
    async def get_depends_result(cls, func: Callable):
        signature = inspect.signature(func)

        kwargs = {}
        for key, arg in signature.parameters.items():
            if not isinstance(arg.default, Depends):
                continue
            kwargs[key] = await cls.recurrent_execute_depends(arg.default)
        return kwargs

    @classmethod
    async def recurrent_execute_depends(cls, depends: Depends):
        func = depends.dependency
        signature = inspect.signature(func)

        if depends.use_cache and func in cls.caches:
            return cls.caches[func]

        kwargs = {}
        for key, arg in signature.parameters.items():
            if arg.default is inspect.Parameter.empty:
                raise E.MachinaException("Dependsにdefault値のない引数を宣言しないでください。")
            if not isinstance(arg.default, Depends):
                kwargs[key] = arg.default
                continue
            kwargs[key] = await cls.recurrent_execute_depends(arg.default)

        if inspect.iscoroutinefunction(func):
            res = await func(**kwargs)
        elif inspect.isasyncgenfunction(func):
            gen = func(**kwargs)
            cls.generators.append(gen)
            res = await gen.__anext__()
        elif inspect.isgeneratorfunction(func):
            gen = func(**kwargs)
            cls.generators.append(gen)
            res = gen.__next__()
        elif inspect.isfunction(func) or inspect.ismethod(func):
            res = func(**kwargs)
        else:
            raise E.MachinaException("funcが想定外のパターンです: {func}")
        if depends.use_cache:
            cls.caches[func] = res
        return res


async def get_depends(func: Callable[..., Any]) -> Any:
    return await DependsContoroller.recurrent_execute_depends(Depends(func))



from __future__ import annotations

from typing import Any, Callable


class Depends:
    def __init__(self, dependency: Callable[..., Any], *, use_cache: bool = True):
        self.dependency = dependency
        self.use_cache = use_cache

    def __repr__(self) -> str:
        attr = getattr(self.dependency, "__name__", type(self.dependency).__name__)
        cache = "" if self.use_cache else ", use_cache=False"
        return f"{self.__class__.__name__}({attr}{cache})"



from __future__ import annotations

import asyncio
import contextvars
import functools
import inspect
import re
from collections import deque
from typing import Any, Callable, Coroutine

_times = dict(d=86400.0, h=3600.0, m=60.0, s=1.0, ms=0.001)
_r = (
    r"^"
    r"(?=.*[dhms(ms)]$)"
    r"((?P<d>\d+(\.\d+)?)(?:d\s*))?"
    r"((?P<h>\d+(\.\d+)?)(?:h\s*))?"
    r"((?P<m>\d+(\.\d+)?)(?:m\s*))?"
    r"((?P<s>\d+(\.\d+)?)(?:s\s*))?"
    r"((?P<ms>\d+)(?:ms\s*))?"
    r"$"
)


def interval_to_second(interval: str) -> float:
    """1s12h35m59s500msのような文字列を秒数に変換する

    Args:
        interval (str): 時間を表す文字列

    Raises:
        ValueError: 変換できなかったときの例外

    Returns:
        float: 秒数
    """
    m = re.match(_r, interval)

    if m is None:
        raise ValueError(f'intervalは[1d12h35m59s500ms]のような形である必要があります。入力: "{interval}"')

    return sum([_times[k] * float(v) for k, v in m.groupdict().items() if v is not None])


async def to_thread(func, *args, **kwargs):
    """Asynchronously run function *func* in a separate thread.

    Any *args and **kwargs supplied for this function are directly passed
    to *func*. Also, the current :class:`contextvars.Context` is propagated,
    allowing context variables from the main thread to be accessed in the
    separate thread.

    Return a coroutine that can be awaited to get the eventual result of *func*.
    """
    loop = asyncio.get_running_loop()
    ctx = contextvars.copy_context()
    func_call = functools.partial(ctx.run, func, *args, **kwargs)
    return await loop.run_in_executor(None, func_call)


class _DummySemaphore:
    async def acquire(self):
        ...

    def release(self):
        ...


class TimeSemaphore:
    def __init__(
        self,
        *,
        entire_calls_limit: int | None = None,
        time_limit: float = 0.0,
        time_calls_limit: int = 1,
    ):
        """通常のSemaphoreに加えて、時間あたりの実行回数制限をかけられるSemaphore

        Args:
            entire_calls_limit (int, optional): 全体の最大並列実行数. Defaults to None.
            time_limit (float, optional): 時間制限[sec]. Defaults to 0.0.
            time_calls_limit (int, optional): 時間制限あたりの最大並列実行数. Defaults to 1.
        """
        self._value = time_calls_limit
        self._time_limit = time_limit
        self.__loop = None
        self._waiters = deque()
        self.entire_calls_limit = entire_calls_limit
        self.__sem = None

    async def __aenter__(self) -> None:
        await self.acquire()

    async def __aexit__(self, *args) -> None:
        self.release()

    @property
    def _loop(self):
        if self.__loop is None:
            self.__loop = asyncio.events.get_event_loop()  # 3.7~
        return self.__loop

    @property
    def _sem(self):
        if self.__sem is None:
            self.__sem = (
                _DummySemaphore() if self.entire_calls_limit is None else asyncio.Semaphore(self.entire_calls_limit)
            )
        return self.__sem

    async def acquire(self) -> bool:
        await self._sem.acquire()

        if self._time_limit == 0.0:
            return True

        while self._value <= 0:
            fut = self._loop.create_future()
            self._waiters.append(fut)
            try:
                await fut
            except BaseException:
                fut.cancel()
                if self._value > 0 and not fut.cancelled():
                    self._value -= 1
                    self._wake_up_next()
                raise
        self._value -= 1
        self._loop.call_later(self._time_limit, self._wake_up_next)
        return True

    def release(self):
        self._sem.release()

    def _wake_up_next(self):
        self._value += 1
        while self._waiters:
            waiter = self._waiters.popleft()
            if not waiter.done():
                waiter.set_result(None)
                return


async def execute_functions(funcs: list[Callable[[], Coroutine[Any, Any, None]] | Callable[[], None]]):
    for func in funcs:
        if inspect.iscoroutinefunction(func):
            await func()  # type: ignore
        elif inspect.isfunction(func) or inspect.ismethod(func):
            func()



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['exmachina', 'exmachina.core', 'exmachina.lib']

package_data = \
{'': ['*']}

install_requires = \
['typing-extensions>=3.10.0,<4.0.0']

extras_require = \
{'rich': ['rich>=10.1.0,<11.0.0']}

setup_kwargs = {
    'name': 'exmachina',
    'version': '0.2.0',
    'description': 'botを作るためのフレームワークです',
    'long_description': '# Ex-Machina\n\n<p align="center">\n  <em>python で bot 書くためのフレームワーク</em>\n</p>\n<p align="center">\n<a href="https://github.com/agarichan/exmachina/actions/workflows/test.yaml" target="_blank">\n  <img src="https://github.com/agarichan/exmachina/actions/workflows/test.yaml/badge.svg?branch=main" alt="Test">\n</a>\n<a href="https://codecov.io/gh/agarichan/exmachina" target="_blank">\n  <img src="https://img.shields.io/codecov/c/github/agarichan/exmachina?color=%2334D058" alt="Coverage">\n</a>\n<a href="https://pypi.org/project/exmachina" target="_blank">\n  <img src="https://img.shields.io/pypi/v/exmachina?color=%2334D058&label=pypi%20package" alt="Package version">\n</a>\n<a href="https://pypi.org/project/exmachina" target="_blank">\n  <img src="https://img.shields.io/pypi/pyversions/exmachina.svg?color=%2334D058" alt="Supported Python versions">\n</a>\n</p>\n\n---\n\n## インストール\n\nusing pip\n\n```\npip install -U exmachina\n```\n\nusing poetry\n\n```\npoetry add exmachina@latest\n```\n\n## 使い方\n\n\n```python\nfrom exmachina import Event, Machina, Retry, RetryFixed\n\nbot = Machina()\n\nbot.create_concurrent_group(\n    name=\'limit\',\n    entire_calls_limit=4,\n    time_calls_limit=3,\n    time_limit=1,\n)\n\n@bot.emit(interval=\'1s\')\nasync def emit(event: Event):\n    res = await event.execute(\'execute\')\n    assert res == 42\n    # or\n    res = await execute()\n    assert res == 42\n\n# 特定の例外でリトライしたい場合\nretry = Retry([\n  RetryFixed(HTTPError, wait_time=1, retries=5),\n])\n\n@bot.execute(concurrent_groups=[\'limit\'], retry=retry)\nasync def execute():\n    return 42\n```\n\n### Emit\n\n定期的に実行したい関数を登録するためのデコレータ\n\n- `name`\n  - Emitでユニークな名前をつける\n  - 省略した場合はデコレートした関数名になる\n- `count`\n  - 実行回数. これで指定した回数実行した後、`alive`が`False`になる\n  - `None`を指定した場合は無限回実行する\n  - デフォルトは`None`\n- `interval`\n  - ループのインターバル `1s`や`1d4h`などと指定できる\n  - デフォルトは`0s`. つまり待機しない\n- `alive`\n  - `True`の場合、botの実行時に自動で実行される\n  - 手動で起動する場合は`False`を指定する\n  - デフォルトは`True`\n\n### Concurrent Group\n\n時間あたりや同時実行数を制限するグループ  \n作成したグループは後述のExecuteに対して設定できる\n\n- `name`\n  - 必須プロパティ. Concurrent Groupでユニークな名前をつける\n- `entire_calls_limit`\n  - 全体の実行数制限\n  - このグループに所属するExecuteの並列実行数\n  - `None`を指定した場合、無制限\n  - デフォルトは`None`\n- `time_calls_limit`\n  - このグループに所属する`time_limit`秒あたりに実行"開始"できるExecuteの数\n  - デフォルトは`1`\n- `time_limit`\n  - `time_calls_limit`の制限時間(秒)\n  - デフォルト`0`. つまり、制限なし\n\n### Execute\n\nEmitから呼び出される、一回きりのタスク  \nEmitは主にbotの制御を行い、Executeは計算処理や外部との通信を行う処理を書く想定\n\n- `name`\n  - Executeでユニークな名前をつける\n  - 省略した場合はデコレートした関数名になる\n- `concurrent_groups`\n  - executeの所属するconcurrent_groupを配列で指定する\n  - デフォルトは`[]`\n- `retry`\n  - Execute実行中に発生した例外をトリガーにリトライを行う設定を指定する\n  - `from exmachina import Retry`\n  - デフォルトは`None`\n\n## Event\n\nEmitする関数に渡されるオブジェクト\n\n- 停止中の別のEmit起動や停止\n- Executeの実行\n- ループの状態を確認できる属性を持つ\n\n### Emitの起動と停止\n\n```python\nevent.start(\'emit_name\')\nevent.stop(\'emit_name\')\n```\n\n### Executeの実行\n\n```python\nevent.execute(\'execute_name\', *args, **kwargs)\n```\n\nまたは、直接呼び出し\n\n```python\nawait execute_name(*args, **kwargs)\n````\n\nevent.executeはexecuteのTaskを返す\n\n### 属性\n\n```python\nevent.epoch # emitのループ回数\nevent.previous_execution_time # 直前のループの処理時間\nevent.count # emitの残りの実行回数(未指定の場合はNone)\n```\n\n## Retry\n\nExecuteのリトライの設定を書くためのもの  \n特定の秒数を待機した後、そのまま再実行を行う\n\n- `rules`\n  - 指定するルールの配列\n- `logger`\n  - 自前のloggerを渡したい場合に使う\n\n### RetryRule\n\n- `RetryFixed`: 常に`wait_time`の間隔でリトライ\n- `RetryFibonacci`: 1,1,2,3,5,...とフィボナッチ数列秒の間隔でリトライ\n- `RetryRange`: `min`と`max`を指定し、その間の乱数秒の間隔でリトライ\n- `RetryExponentialAndJitter`: 指数関数倍的に最大の待機時間を伸ばしつつリトライ\nのいずれかを使用する\n\n共通引数\n- `exception`\n  - 指定する例外のクラス\n- `retries`\n  - リトライ回数. 省略すると`RecursionError`が出るまで再実行する\n  - デフォルトは`None`\n- `filter`\n  - 例外インスタンスを引数に`bool`を返す関数を指定\n  - Trueを返した場合にこのリトライ条件にマッチする\n  - HTTPのステータスコードなどで引っ掛けたいリトライ設定が異なる場合などを想定\n  - デフォルトは`None`. つまり、常に`True`を返す\n## 開発\n\n### init\n\n```bash\npoetry install\npoetry shell\n```\n\n### fmt\n\n```\npoe fmt\n```\n\n### lint\n\n```\npoe lint\n```\n\n### test\n\n```\npoe test\n```\n',
    'author': 'agarichan',
    'author_email': None,
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/agarichan/exmachina',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



from __future__ import annotations

import asyncio
import functools
import inspect
import logging
from collections import defaultdict
from dataclasses import InitVar, dataclass, field
from functools import partial
from time import perf_counter
from typing import Any, Awaitable, Callable, Coroutine, NoReturn, TypeVar

from exmachina.lib.helper import TimeSemaphore, execute_functions, interval_to_second

from . import exception as E
from .depends_contoroller import DependsContoroller
from .helper import set_verbose
from .retry import Retry

try:
    from typing import Literal  # type: ignore
except ImportError:
    from typing_extensions import Literal

DecoratedCallable = TypeVar("DecoratedCallable", bound=Callable[..., Awaitable[None]])
DecoratedResultCallable = TypeVar("DecoratedResultCallable", bound=Callable[..., Awaitable[Any]])


@dataclass
class Emit:
    name: str
    func: Callable[..., Awaitable[None]]
    interval: str  # 次回の実行までの待機時間
    mode: Literal["after", "entire"]
    alive: bool  # 動作中か非動作中か
    count: int | None = None  # ループの回数、未指定の場合無限回 (immutable)


@dataclass
class Execute:
    name: str
    func: Callable[..., Awaitable[Any]]
    concurrent_groups: list[ConcurrentGroup] = field(default_factory=list)
    retry: Retry | None = None


@dataclass
class ConcurrentGroup:
    name: str
    semaphore: TimeSemaphore


@dataclass
class Event:
    epoch: int  # 1,2,...
    previous_execution_time: float  # seconds

    bot: InitVar[Machina]

    def __post_init__(self, bot: Machina):
        self._bot = bot

    def stop(self, emit_name: str, force: bool = False):
        if force:
            task = self._bot._emit_tasks[emit_name]
            task.cancel()
        else:
            self._bot._emits[emit_name].alive = False

    def start(self, emit_name: str) -> asyncio.Task:
        self._bot._add_emit_task(emit_name)
        return self._bot._emit_tasks[emit_name]

    def get(self, emit_name: str) -> asyncio.Task | None:
        return self._bot._emit_tasks.get(emit_name)

    def execute(self, execute_name: str, *args, **kwargs) -> asyncio.Task:
        task = self._bot._add_execute_task(execute_name, *args, **kwargs)
        return task


class TaskManager:
    emit_tasks: list[asyncio.Task[None]]
    execute_tasks: list[asyncio.Task]


class Machina:
    def __init__(
        self,
        *,
        on_startup: list[Callable[[], Coroutine[Any, Any, None]] | Callable[[], None]] = [],
        on_shutdown: list[Callable[[], Coroutine[Any, Any, None]] | Callable[[], None]] = [],
        logger: logging.Logger | None = None,
        verbose: Literal["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"] | None = None,
    ) -> None:
        self._emits: dict[str, Emit] = {}
        self._executes: dict[str, Execute] = {}
        self._concurrent_groups: dict[str, ConcurrentGroup] = {}
        self._emit_tasks: dict[str, asyncio.Task] = {}
        self._execute_tasks: dict[str, list[asyncio.Task]] = defaultdict(list)
        self._execute_task_executings: dict[str, int] = defaultdict(int)
        self.on_startup = on_startup
        self.on_shutdown = on_shutdown
        self.logger = logger or logging.getLogger(__name__)
        # 全てのタスクが終わったことを確認するようの変数
        self._unfinished_tasks = 0
        self.__finished = None

        if verbose is not None:
            set_verbose(logger, verbose)

    @property
    def _finished(self):
        """loop作成前にEventは作成できないので実行時に取得する"""
        if self.__finished is None:
            self.__finished = asyncio.Event()
        return self.__finished

    async def run(self):
        """botを起動する関数
        全てのemitが停止するまで永遠に待機する
        """
        await self._startup()

        try:
            for emit in self._emits.values():
                if emit.alive:
                    self._add_emit_task(emit.name)

            if self._unfinished_tasks > 0:
                self._finished.clear()
                await self._finished.wait()
                for task in self._emit_tasks.values():
                    await task
        finally:
            await self._shutdown()

    async def _startup(self):
        self.__finished = None
        await execute_functions(self.on_startup)

    async def _shutdown(self):
        await execute_functions(self.on_shutdown)

    def create_concurrent_group(
        self, name: str, entire_calls_limit: int | None = None, time_limit: float = 0, time_calls_limit: int = 1
    ) -> ConcurrentGroup:
        """並列実行の制限グループを作成する

        例えば、1秒あたりに5回まで実行を許可し、実行中タスクの最大数を6個にする場合は
        time_limit = 1, time_calls_limit = 5, entire_calls_limit = 6
        とする

        Args:
            name (str): 名前
            entire_calls_limit (int, optional): グループ全体での最大並列実行数. Defaults to None.
            time_limit (float, optional): 制限時間[sec]. Defaults to 0.
            time_calls_limit (int, optional): 制限時間あたりの最大並列実行数. Defaults to 1.

        Raises:
            E.MachinaException: 同じ名前を登録しようとした時の例外

        Returns:
            ConcurrentGroup: 並列実行の制限グループ
        """
        if name in self._concurrent_groups:
            raise E.MachinaException(f"このconcurrent_groupはすでに登録されています別の名前にしてください: [{name}]")

        cg = ConcurrentGroup(
            name=name,
            semaphore=TimeSemaphore(
                entire_calls_limit=entire_calls_limit, time_limit=time_limit, time_calls_limit=time_calls_limit
            ),
        )
        self._concurrent_groups[name] = cg
        return cg

    def emit(
        self,
        *,
        name: str | None = None,
        count: int | None = None,
        interval: str = "0s",
        mode: Literal["after", "entire"] = "after",
        alive: bool = True,
    ) -> Callable[[Callable[..., Awaitable[None]]], Callable[[], NoReturn]]:
        """Emitを登録します

        Emitは基本的にループで一定間隔で登録された関数を繰り返します
        登録する関数はevent: Eventを引数に取れます

        Args:
            name (str, optional): 名前(ユニーク),省略するとデコレートした関数名を使用する
            count (int, optional): 指定すると指定回数だけループが回ったあと終了する. Defaults to None.
            interval (str, optional): 次の実行までの待機時間. Defaults to "0s".
            mode (Literal['after', 'entire'], optional): 待機時間に処理時間を含むかどうか. Defaults to "after".
                after: 処理の後interval秒待機する
                entire: intervalから処理時間を引いた時間待機する
            alive (bool, optional): ループを稼働するかどうか. Defaults to True.
        """
        if count is not None and count < 0:
            raise E.MachinaException("countは0以上を指定してください")

        def decorator(func: Callable[..., Awaitable[None]]):
            _name = func.__name__ if name is None else name
            emit = Emit(name=_name, func=func, interval=interval, alive=alive, mode=mode, count=count)
            if _name in self._emits:
                raise E.MachinaException(f"このemitはすでには登録されています、別の名前にしてください: [{_name}]")
            self._emits[_name] = emit

            def no_return():
                raise E.MachinaException("emitは直接呼び出せません")

            return no_return

        return decorator

    def execute(self, *, name: str | None = None, concurrent_groups: list[str] = [], retry: Retry | None = None):
        def decorator(func: DecoratedResultCallable) -> DecoratedResultCallable:
            _name = func.__name__ if name is None else name
            _concurrent_groups = [self._concurrent_groups[cg_name] for cg_name in concurrent_groups]
            execute = Execute(name=_name, func=func, concurrent_groups=_concurrent_groups, retry=retry)
            if _name in self._executes:
                raise E.MachinaException(f"このexecuteはすでには登録されています、別の名前にしてください: [{_name}]")
            self._executes[_name] = execute

            @functools.wraps(func)
            def wrap(*args, **kwargs):
                task = self._add_execute_task(_name, *args, **kwargs)
                return task

            return wrap  # type: ignore

        return decorator

    def _add_emit_task(self, name: str) -> str:
        emit = self._emits.get(name)
        if emit is None:
            raise E.MachinaException(f"emitsに存在しないnameを指定しています: [{name}]")

        task = self._emit_tasks.get(name)
        if task is not None:
            if not task.done():
                self.logger.warning(f"emit taskを起動しようとしましたが、すでに作動中です: [{name}]")
                return name

        emit.alive = True

        @cancelled_wrapper(name, "emit", self.logger)
        async def func():
            return await set_interval(emit, self)

        task = asyncio.create_task(func())
        task.add_done_callback(partial(self._emit_task_done, emit))
        self._emit_tasks[emit.name] = task
        self._unfinished_tasks += 1

        return name

    def _emit_task_done(self, emit: Emit, _: asyncio.Task):
        """emitのtaskが終了した時(cancelも含む)に呼ばれるcallback関数
        タスクが全て終わったことを確認するための変数を更新する
        """
        self._unfinished_tasks -= 1
        emit.alive = False
        self.logger.debug(f'Done emit task: "{emit.name}", remain tasks: {self._unfinished_tasks}')
        if self._unfinished_tasks == 0:
            self._finished.set()

    def _add_execute_task(self, name: str, *args, **kwargs) -> asyncio.Task:
        execute = self._executes.get(name)
        if execute is None:
            raise E.MachinaException(f"executesに存在しないnameを指定しています: [{name}]")

        tasks = self._execute_tasks[name]

        @cancelled_wrapper(name, "execute", self.logger)
        async def func():
            return await execute_wrapper(execute, self, *args, **kwargs)

        task = asyncio.create_task(func())
        task.add_done_callback(partial(self._execute_task_done, execute))
        tasks.append(task)

        self._unfinished_tasks += 1

        return task

    def _execute_task_done(self, execute: Execute, task: asyncio.Task):
        self._unfinished_tasks -= 1
        tasks = self._execute_tasks[execute.name]
        # 終了したこのタスク自身をリストから消去
        tasks.pop(tasks.index(task))

        if self._unfinished_tasks == 0:
            self._finished.set()


def cancelled_wrapper(name: str, type: Literal["emit", "execute"], logger: logging.Logger):
    def decorator(afunc: Callable[[], Awaitable[Any]]):
        @functools.wraps(afunc)
        async def _inner():
            try:
                return await afunc()
            except asyncio.CancelledError:
                logger.debug(f'Cancelled {type} task: "{name}"')
            except BaseException:
                logger.error(f'Uncatched error {type} task: "{name}"', exc_info=True)
                raise

        return _inner

    return decorator


def get_args(func) -> tuple[list[str], dict[str, Any]]:
    signature = inspect.signature(func)
    args = [k for k, v in signature.parameters.items() if v.default is inspect.Parameter.empty]
    kwargs = {k: v.default for k, v in signature.parameters.items() if v.default is not inspect.Parameter.empty}
    return args, kwargs


async def set_interval(emit: Emit, bot: Machina):
    bot.logger.debug(f'Start emit task: "{emit.name}"')
    interval = interval_to_second(emit.interval)
    previous_execution_time = 0.0
    epoch = 1
    count = emit.count
    while count is None or count > 0:
        # デバッグ用の変数
        event = Event(
            epoch=epoch,
            previous_execution_time=previous_execution_time,
            bot=bot,
        )
        # 引数の設定
        args, kwargs = get_args(emit.func)
        if "event" in args:
            kwargs.update({"event": event})
        # Dependsの実行
        kwargs.update(await DependsContoroller.get_depends_result(emit.func))
        start = perf_counter()
        # 実行
        await emit.func(**kwargs)
        # 計測
        previous_execution_time = perf_counter() - start
        epoch += 1
        if count is not None:
            count -= 1
            if count <= 0:
                break
        if not emit.alive:
            break
        # 待機
        if emit.mode == "after":
            await asyncio.sleep(interval)
        else:
            wait = interval - previous_execution_time
            if wait < 0:
                bot.logger.warning(f"指定されたインターバルより{-wait:.0f}秒以上遅延しています.")
            await asyncio.sleep(0 if wait < 0 else wait)


async def execute_wrapper(execute: Execute, bot: Machina, *args, **kwargs):

    retry = execute.retry or (lambda func: func)

    @retry
    async def nest(concurrent_groups: list[ConcurrentGroup]):
        if concurrent_groups != []:
            async with concurrent_groups[0].semaphore:
                return await nest(concurrent_groups[1:])
        else:
            bot._execute_task_executings[execute.name] += 1
            _t = len(bot._execute_tasks[execute.name])
            _e = bot._execute_task_executings[execute.name]
            bot.logger.debug(f'Start execute task: "{execute.name}" [tasks={_t}, executings={_e}]')
            # Dependsの実行
            kwargs.update(await DependsContoroller.get_depends_result(execute.func))
            res = await execute.func(*args, **kwargs)
            bot._execute_task_executings[execute.name] -= 1
            return res

    return await nest(execute.concurrent_groups)
